#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <limits.h>
#include <stdint.h>
#include <errno.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "xmalloc.h"
#include "opcode.h"
#include "misc.h"
#include "buffer.h"
#include "obj.h"

#ifdef _MEM_AREA
#include "mem_area.h"
#endif

#define err_msg(fmt, arg...)	\
do {	\
	message(1, file_name, line_number, fmt, ##arg); \
} while(0)

extern char		*__progname;
static char		*file_name;
static unsigned int	line_number;

struct label {
	int	ln;
	int	addr;
};

/*
 * `labels' contains addresses (offsets in the executable section)  when
 *  we met this label.
 *
 * `need_patch' contains addresses where we'll write a patch value.
 *
 * The patch value is relative about %pc and it is the difference between
 * two `addr' so it can be negative and positive. See backpatch_labels()
 * function.
 *
 */
struct {
	struct label *lb;
	unsigned int used;
	unsigned int size;
} labels, need_patch;

typedef int (*instr_handler_t)(char *cur);
/* Handler prototypes. Handlers return 0 in case of success and a negative value
 * in case of fault. */
static int data_sect(char *cur);
static int init_sect(char *cur);
static int text_sect(char *cur);

static int jump(char *cur);
static int arithm(char *cur);
static int libcall(char *cur);
static int label(char *cur);
static int load(char *cur);
static int store(char *cur);
static int call(char *cur);
static int stack_op(char *cur);
static int func_op(char *cur);
static int cmp(char *cur);
static int push(char *cur);
static int strpush(char *cur);
static int refer(char *cur);
static int ainit(char *cur);
static int declaration(char *cur);
static int global(char *cur);
static int size_op(char *cur);

/* The table of instructions handlers. */
struct code_entry {
	const char	*name;
	instr_handler_t	handler;
} instr_handlers[] = {
	{ ".init",		init_sect },
	{ ".text",		text_sect },
	{ ".data",		data_sect },
	{ ".globl",		global },
	{ ".size",		size_op },
	{ "add",		arithm },
	{ "sub",		arithm },
	{ "mul",		arithm },
	{ "div",		arithm },
	{ "mod",		arithm },
	{ "fadd",		arithm },
	{ "fsub",		arithm },
	{ "fmul",		arithm },
	{ "fdiv",		arithm },
	{ "fmod",		arithm },
	{ "cmp",		cmp },
	{ "fcmp",		cmp },
	{ "strcmp",		cmp },
	{ "libcall",		libcall },
	{ "ent",		func_op },
	{ "leave",		func_op },
	{ "call",		call },
	/* Jumps. */
	{ "jg",			jump },
	{ "jge",		jump },
	{ "jl",			jump },
	{ "jle",		jump },
	{ "je",			jump },
	{ "jne",		jump },
	{ "jmp",		jump },
	{ ".L",			label },
	{ "push",		push },
	{ "strpush",		strpush },
	/* Load instructions. */
	{ "load",		load },
	{ "aload",		load },
	{ "strload",		load },
	{ "astrload",		load },
	/* Store instructions. */
	{ "store",		store },
	{ "store0",		store },
	{ "astore",		store },
	{ "astore0",		store },
	{ "strstore",		store },
	{ "strstore0",		store },
	{ "astrstore",		store },
	{ "astrstore0",		store },
	/* Reference instructions. */
	{ "refput",		refer },
	{ "refcopy",		refer },
	{ "refload",		refer },
	{ "refstore",		refer },
	{ "refstore0",		refer },
	{ "refstrload",		refer },
	{ "refstrstore",	refer },
	{ "refstrstore0",	refer },
	/* Array init size. */
	{ "ainit",		ainit },
	{ "grow",		stack_op },
	{ "shrink",		stack_op },
	{ NULL,			declaration }
};

/* Pointer to text or init buffer. */
static struct buffer *cur_text_sect;
/* Buffers to store text section and init section.
 * We code all data in the big endian format. */
static struct buffer *text_be_sect;
static struct buffer *init_be_sect;

/* Buffers to store read only, string, relocation table sections.
 * We also need buffers to store symbol table sections in the current endian
 * format and in the big endian format. */
static struct buffer *ro_sect;
static struct buffer *string_sect;
static struct buffer *reloc_be_sect;
static struct buffer *symbol_be_sect;
static struct buffer *symbol_sect;
/* We don't use buffer for a data section we only need to calculate its size
 * for runtime. */
static unsigned int data_section_size;
/* Last index in the symbol table. */
static unsigned int last_symbol_idx;

/*static int str_to_uint(const char *str, char **endptr ,unsigned int *num)
{
        unsigned int n;
	char *end;

        errno = 0;
        n = strtoul(str, &end, 0);

        if ((errno == ERANGE && (n == 0 || n == ULONG_MAX)) || 
                        errno == EINVAL || errno != 0)
                return -1;

        *num = n;

	if (endptr != NULL && *endptr != NULL)
		*endptr = end;

        return 0;
}*/

static void patch_later(struct label lb)
{
	if (need_patch.used == need_patch.size) {
		need_patch.size += 16;
		need_patch.lb = xrealloc(need_patch.lb, need_patch.size*sizeof(lb));
	}

	need_patch.lb[need_patch.used++] = lb;
}

static void add_label(struct label lb)
{
	if (labels.used == labels.size) {
		labels.size += 16;
		labels.lb = xrealloc(labels.lb, labels.size*sizeof(lb));
	}

	labels.lb[labels.used++] = lb;
}

/* We don't put the string as it is, first we try to find if the string was
 * already added to this buffer earlier. */
static unsigned int put_string(struct buffer *buffer, const char *str)
{
	unsigned int ret;
	char *c, *end;
	/* If the buffer is empty, first off all we add the empty string. */
	if (buffer_size(buffer) == 0)
		buffer_put_string(string_sect, "");

	end = buffer_ptr(buffer) + buffer_size(buffer);
	c = buffer_ptr(buffer);

	while (c < end) {
		if (!strcmp(c, str))
			return (c - (char *)buffer_ptr(buffer));
		c += strlen(c);
		c++;
	}

	ret = buffer_size(buffer);
	buffer_put_string(buffer, str);

	return ret;
}

/* Make a look up in the symbol table. And return the pointer to the symbol. */
static struct sym_tb_entry *symbol_table_lookup(const char *name)
{
	struct sym_tb_entry *sym_table;
	unsigned int table_size, i;

	sym_table = (struct sym_tb_entry *)buffer_ptr(symbol_sect);
	table_size = buffer_size(symbol_sect) / sizeof(struct sym_tb_entry);

	for (i = 0; i < table_size; i++) {
		char *cmp_name = buffer_ptr(string_sect) + sym_table[i].st_name;
		if (!strcmp(name, cmp_name))
			return &sym_table[i];
	}

	return NULL;
}

/* Make a look up in the symbol table but return the symbol order number in
 * the symbol table. */
static unsigned int get_symbol_index(const char *name)
{
	struct sym_tb_entry *sym_table;
	unsigned int table_size, i;

	sym_table = (struct sym_tb_entry *)buffer_ptr(symbol_sect);
	table_size = buffer_size(symbol_sect)/sizeof(struct sym_tb_entry);

	for (i = 0; i < table_size; i++) {
		char *cmp_name = (char *)buffer_ptr(string_sect) + sym_table[i].st_name;
		if (!strcmp(name, cmp_name))
			return i;
	}

	return 0;
}

static void skip_blanks(char **cur)
{
	char *pos;

	pos = *cur;
	while (isblank(*pos))
		++pos;

	*cur = pos;
}

static void skip_id(char **cur)
{
	char *pos;

	pos = *cur;

	if (isalpha(*pos) || *pos == '_') {
		do {
			++pos;
		} while (isalnum(*pos) || *pos == '_');
	}

	*cur = pos;
}

static int data_sect(char *cur)
{
	return 0;
}

static int text_sect(char *cur)
{
	cur_text_sect = text_be_sect;
	return 0;
}

static int init_sect(char *cur)
{
	cur_text_sect = init_be_sect;
	return 0;
}

static sh_type_t section_type()
{
	if (cur_text_sect == NULL)
		return SHT_DATA;

	if (cur_text_sect == init_be_sect)
		return SHT_INIT;

	return SHT_TEXT;
}

static int global(char *cur)
{
	struct sym_tb_entry *s, s1;
	char *pos0;

	/* Skip a dot.*/
	++cur;

	skip_id(&cur);
	skip_blanks(&cur);

	if (*cur == 0) {
		err_msg("wait name id");
		return -1;
	}

	pos0 = cur;
	skip_id(&cur);

	*cur = 0;
	s = symbol_table_lookup(pos0);
	/* In case of handling function, the symbol could be added before when
	 * the function was called from another place with `call' instruction
	 * and in this case we need only fix some fields. */
	if (s != NULL) {
		s->st_visib = ST_GLOBAL;
		s->st_shtype = section_type();
	} else {
		memset(&s1, 0, sizeof(struct sym_tb_entry));
		s1.st_name = put_string(string_sect, pos0);
		s1.st_visib = ST_GLOBAL;
		s1.st_shtype = section_type();
		obj_put_sym_tb_entry(symbol_sect, &s1);
	}

	return 0;
}

static int jump(char *cur)
{
	struct label lb;
	char *pos0, ch;
	unsigned int ln;
	int ret;

	return_val_if_fail(cur_text_sect != NULL, -1);

	pos0 = cur;
	skip_id(&cur);
	ch = *cur;
	*cur = 0;

	if (STREQ("jg", pos0)) {
		buffer_put_u8(cur_text_sect, JG);
	} else if (STREQ("jge", pos0)) {
		buffer_put_u8(cur_text_sect, JGE);
	} else if (STREQ("jl", pos0)) {
		buffer_put_u8(cur_text_sect, JL);
	} else if (STREQ("jle", pos0)) {
		buffer_put_u8(cur_text_sect, JLE);
	} else if (STREQ("je", pos0)) {
		buffer_put_u8(cur_text_sect, JEQ);
	} else if (STREQ("jne", pos0)) {
		buffer_put_u8(cur_text_sect, JNEQ);
	} else if (STREQ("jmp", pos0)) {
		buffer_put_u8(cur_text_sect, JMP);
	} else {
		err_msg("unknown a jump instruction `%s'", pos0);
		return -1;
	}

	*cur = ch;
	skip_blanks(&cur);

	if (*cur == 0) {
		err_msg("jump instruction waits a label");
		return -1;
	}

	if (cur[0] != '.' || cur[1] != 'L') {
		err_msg("wrong jump label format");
		return -1;
	}

	cur += 2;

	ret = str_to_uint(cur, &cur, &ln, 0);
	if (ret < 0) {
		err_msg("label number %d", ln);
		return -1;
	}

	lb.ln = ln;
	lb.addr = buffer_size(cur_text_sect);

	patch_later(lb);
	/* Put zero, we will patch it lately. */
	buffer_put_u32_be(cur_text_sect, 0);

	return 0;
}

static int label(char *cur)
{
	struct label lb;
	unsigned int ln;
	int ret;

	cur += 2; /* skip .L */

	ret = str_to_uint(cur, &cur, &ln, 0);
	if (ret < 0) {
		err_msg("label number %d", ln);
		return -1;
	}

	lb.ln = ln;
	lb.addr = buffer_size(cur_text_sect);
	add_label(lb);

	return 0;
}

static int size_op(char *cur)
{
	struct sym_tb_entry *s, *sym_table;
	unsigned int size;
	char *name;
	int ret;

	return_val_if_fail(string_sect != NULL, -1);

	sym_table = (struct sym_tb_entry *)buffer_ptr(symbol_sect);
	s = &sym_table[last_symbol_idx];

	name = buffer_ptr(string_sect) + s->st_name;

	++cur;

	skip_id(&cur);
	skip_blanks(&cur);

	size = 0;
	switch (s->st_shtype) {
	case SHT_DATA:
		ret = str_to_uint(cur, &cur, &size, 0);
		if (ret < 0) {
			err_msg("size %d", size);
			return -1;
		}
		data_section_size += size;
		break;
	case SHT_TEXT:
		size = buffer_size(cur_text_sect) - s->st_offset;
		break;
	default:
		err_msg("unknown section header type");
		return -1;
	}

	s->st_size = size;

	return 0;
}
/* Indeed a declaration looks like a simple label `:label_name'. */
static int declaration(char *cur)
{
	struct sym_tb_entry s1, *s;
	char *pos0;

	pos0 = cur;

	skip_id(&cur);

	if (*cur != ':') {
		err_msg("wrong declaration");
		return -1;
	}

	*cur = 0;
	s = symbol_table_lookup(pos0);

	if (s == NULL) {
		memset(&s1, 0, sizeof(struct sym_tb_entry));
		s1.st_name = put_string(string_sect, pos0);
		s1.st_visib = ST_MODULE;
		s1.st_shtype = section_type();
		obj_put_sym_tb_entry(symbol_sect, &s1);
		s = symbol_table_lookup(pos0);
	}

	switch (s->st_shtype) {
	case SHT_DATA:
		s->st_offset = data_section_size;
		break;
	case SHT_TEXT:
		s->st_offset = buffer_size(cur_text_sect);
		break;
	default:
		err_msg("unknown section header type");
		return -1;
	}

	last_symbol_idx = get_symbol_index(pos0);

	return 0;
}

static int arithm(char *cur)
{
	char *pos0;

	return_val_if_fail(cur_text_sect != NULL, -1);

	pos0 = cur;
	skip_id(&cur);
	*cur = 0;

	if (STREQ("add", pos0)) {
		buffer_put_u8(cur_text_sect, ADD);
	} else if (STREQ("sub", pos0)) {
		buffer_put_u8(cur_text_sect, SUB);
	} else if (STREQ("mul", pos0)) {
		buffer_put_u8(cur_text_sect, MUL);
	} else if (STREQ("div", pos0)) {
		buffer_put_u8(cur_text_sect, DIV);
	} else if (STREQ("mod", pos0)) {
		buffer_put_u8(cur_text_sect, MOD);
	} else if (STREQ("fadd", pos0)) {
		buffer_put_u8(cur_text_sect, FADD);
	} else if (STREQ("fsub", pos0)) {
		buffer_put_u8(cur_text_sect, FSUB);
	} else if (STREQ("fmul", pos0)) {
		buffer_put_u8(cur_text_sect, FMUL);
	} else if (STREQ("fdiv", pos0)) {
		buffer_put_u8(cur_text_sect, FDIV);
	} else if (STREQ("fmod", pos0)) {
		buffer_put_u8(cur_text_sect, FMOD);
	} else {
		err_msg("unknown arithmetic instruction `%s'", pos0);
		return -1;
	}

	return 0;
}

static int stack_op(char *cur)
{
	char *pos0, ch;
	unsigned int size;
	int ret;

	return_val_if_fail(cur_text_sect != NULL, -1);

	pos0 = cur;
	skip_id(&cur);
	ch = *cur;
	*cur = 0;

	if (STREQ("grow", pos0)) {
		buffer_put_u8(cur_text_sect, GROW);
	} else if (STREQ("shrink", pos0)) {
		buffer_put_u8(cur_text_sect, SHRINK);
	} else {
		err_msg("unknown stack instruction `%s'", pos0);
		return -1;
	}

	*cur = ch;
	skip_blanks(&cur);

	if (*cur != '$') {
		err_msg("stack instruction operand");
		return -1;
	}

	++cur;
	ret = str_to_uint(cur, &cur, &size, 0);
	if (ret < 0) {
		err_msg("stack operand %d", size);
		return -1;
	}

	buffer_put_u32_be(cur_text_sect, size);

	return 0;
}

static int call(char *cur)
{
	struct sym_tb_entry *s;
	struct rel_tb_entry r;
	char *pos0;

	/* Skip `call' instruction. */
	skip_id(&cur);
	skip_blanks(&cur);

	buffer_put_u8(cur_text_sect, CALL);
	/* Skip a function name. */
	pos0 = cur;
	skip_id(&cur);

	*cur = 0;

	s = symbol_table_lookup(pos0);
	if (s == NULL) {
		struct sym_tb_entry s1;
		memset(&s1, 0, sizeof(struct sym_tb_entry));
		s1.st_name = put_string(string_sect, pos0);
		s1.st_shtype = SHT_UNDEF;
		s1.st_visib = ST_GLOBAL;
		obj_put_sym_tb_entry(symbol_sect, &s1);
	}

	r.r_info = get_symbol_index(pos0);
	r.r_flag = RF_SYMTB;
	r.r_offset = buffer_size(cur_text_sect);
	r.r_shtype = section_type();

	buffer_put_u32_be(cur_text_sect, 0);
	obj_put_reloc_tb_entry_be(reloc_be_sect, &r);

	return 0;
}

static int cmp(char *cur)
{
	char *pos0;

	return_val_if_fail(cur_text_sect != NULL, -1);

	pos0 = cur;
	skip_id(&cur);
	*cur = 0;

	if (STREQ("cmp", pos0)) {
		buffer_put_u8(cur_text_sect, CMP);
	} else if (STREQ("fcmp", pos0)) {
		buffer_put_u8(cur_text_sect, FCMP);
	} else if (STREQ("strcmp", pos0)) {
		buffer_put_u8(cur_text_sect, STRCMP);
	} else {
		err_msg("unknown instruction `%s'", pos0);
		return -1;
	}

	return 0;
}

static int func_op(char *cur)
{
	char *pos0;

	return_val_if_fail(cur_text_sect != NULL, -1);

	pos0 = cur;
	skip_id(&cur);
	*cur = 0;

	if (STREQ("ent", pos0)) {
		buffer_put_u8(cur_text_sect, ENT);
	} else if (STREQ("leave", pos0)) {
		buffer_put_u8(cur_text_sect, LEAVE);
	} else {
		err_msg("unknown instruction `%s'", pos0);
		return -1;
	}

	return 0;
}


static int libcall(char *cur)
{
	unsigned int calln;
	int ret;

	return_val_if_fail(cur_text_sect != NULL, -1);

	skip_id(&cur);
	skip_blanks(&cur);

	if (*cur != '$') {
		err_msg("library call number");
		return -1;
	}

	++cur;
	buffer_put_u8(cur_text_sect, LIBCALL);


	ret = str_to_uint(cur, &cur, &calln, 0);
	if (ret < 0) {
		err_msg("library call number %d", calln);
		return -1;
	}

	buffer_put_u32_be(cur_text_sect, calln);

	return 0;
}

static int push(char *cur)
{
	unsigned int num;
	int ret;

	return_val_if_fail(cur_text_sect != NULL, -1);

	buffer_put_u8(cur_text_sect, PUSH);
	/* Skip `push' instruction and blanks after. */
	skip_id(&cur);
	skip_blanks(&cur);

	if (*cur == 0) {
		err_msg("push waits operand");
		return -1;
	}

	ret = str_to_uint(cur, &cur, &num, 0);
	if (ret < 0) {
		err_msg("push operand %d", num);
		return -1;
	}

	buffer_put_u32_be(cur_text_sect, num);

	return 0;
}

static int strpush(char *cur)
{
	struct rel_tb_entry r;
	unsigned int offs;
	char *ch;

	return_val_if_fail(cur_text_sect != NULL, -1);

	buffer_put_u8(cur_text_sect, STRPUSH);
	/* Skip a `push' instruction and blanks after. */
	skip_id(&cur);
	skip_blanks(&cur);

	if (*cur == 0) {
		err_msg("need a string constant");
		return -1;
	}

	if (*cur != '\"') {
		err_msg("not a string constant");
		return -1;
	}

	++cur;
	ch = strchr(cur, '\"');
	if (ch == NULL) {
		err_msg("not a string constant");
		return -1;
	}

	*ch = 0;
	r.r_info = SHT_RO;
	r.r_offset = buffer_size(cur_text_sect);
	r.r_flag = RF_SECT;
	r.r_shtype = section_type();

	obj_put_reloc_tb_entry_be(reloc_be_sect, &r);

	offs = put_string(ro_sect, cur);
	buffer_put_u32_be(cur_text_sect, offs);

	return 0;
}

/* The most important routine, we decode the instruction operand, it can address
 * the stack, a memory area or the data section. */
static int handle_operand(char *cur)
{
	struct sym_tb_entry *s;
	struct rel_tb_entry r;
	unsigned int off, nbit;
	char *pos0;
	int neg, ret, _off;

	skip_blanks(&cur);
	/* Stack operand. */
	if (cur[0] == '-' || cur[0] == '$') {
		neg = 0;
		if (cur[0] == '-')
			neg = 1;
		ret = str_to_uint(cur+neg+1, &cur, &off, 0);
		if (ret < 0) {
			err_msg("stack offset %d", off);
			return -1;
		}

		_off = neg == 1 ? -off : off;
		buffer_put_u8(cur_text_sect, FP_OFFS);
		buffer_put_u32_be(cur_text_sect, _off);
	}
#ifdef _MEM_AREA
	else if (cur[0] == '#') {
		unsigned int res;
		unsigned char type_off;

		res = 0;
		++cur;
		switch (*cur) {
		case 'I':
			type_off = IAREA_OFFS;
			break;
		case 'Q':
			type_off = QAREA_OFFS;
			break;
		default:
			err_msg("not I or Q area");
			return -1;
		}

		++cur;
		
		ret = str_to_uint(cur, &cur, &off, 0);
		if (ret < 0) {
			err_msg("memory area offset %d", off);
			return -1;
		}

		res = SET_OFFSET(res, off);
		switch (cur[0]) {
		case 'U':
			switch (cur[1]) {
			case 'B':
				res = SET_DATA_TYPE(res, UBYTE);
				break;
			case 'S':
				res = SET_DATA_TYPE(res, USHORT);
				break;
			case 'L':
				res = SET_DATA_TYPE(res, ULONG);
				break;
			default:
				err_msg("not UB, US or UL data type");
				return -1;
			}
			++cur;
			break;
		case 'B':
			res = SET_DATA_TYPE(res, BYTE);
			break;
		case 'S':
			res = SET_DATA_TYPE(res, SHORT);
			break;
		case 'L':
			res = SET_DATA_TYPE(res, LONG);
			break;
		case 'F':
			res = SET_DATA_TYPE(res, FLOAT);
			break;
		default:
			err_msg("not B, S, L or F data type");
			return -1;
		}

		++cur;

		if (cur[0] == '.' && cur[1] == 'B') {
			res = BIT_PRESENT(res);
			cur += 2;
			ret = str_to_uint(cur, &cur, &nbit, 0);
			if (ret < 0) {
				err_msg("bit number out of the range %d", nbit);
				return -1;
			}
			res = SET_BIT(res, nbit);
		}
		buffer_put_u8(cur_text_sect, type_off);
		buffer_put_u32_be(cur_text_sect, res);
	}
#endif
	else {
		pos0 = cur;

		skip_id(&cur);
		*cur = 0;

		buffer_put_u8(cur_text_sect, DATA_OFFS);
		memset(&r, 0, sizeof(struct rel_tb_entry));

		s = symbol_table_lookup(pos0);

		if (s == NULL) {
			struct sym_tb_entry s1;
			memset(&s1, 0, sizeof(struct sym_tb_entry));
			s1.st_name = put_string(string_sect, pos0);
			s1.st_shtype = SHT_UNDEF;
			s1.st_visib = ST_GLOBAL;
			obj_put_sym_tb_entry(symbol_sect, &s1);
		}

		r.r_info = get_symbol_index(pos0);
		r.r_flag = RF_SYMTB;
		r.r_offset = buffer_size(cur_text_sect);
		r.r_shtype = section_type();

		buffer_put_u32_be(cur_text_sect, 0);
		obj_put_reloc_tb_entry_be(reloc_be_sect, &r);
	}

	return 0;
}

static int load(char *cur)
{
	char *pos0, ch;
	int res;

	return_val_if_fail(cur_text_sect != NULL, -1);

	pos0 = cur;
	skip_id(&cur);

	ch = *cur;
	*cur = 0;

	if (STREQ("load", pos0)) {
		buffer_put_u8(cur_text_sect, LOAD);
	} else if (STREQ("aload", pos0)) {
		buffer_put_u8(cur_text_sect, ALOAD);
	} else if (STREQ("strload", pos0)) {
		buffer_put_u8(cur_text_sect, STRLOAD);
	} else if (STREQ("astrload", pos0)) {
		buffer_put_u8(cur_text_sect, ASTRLOAD);
	} else {
		err_msg("unknown load instruction `%s'", pos0);
		return -1;
	}

	*cur = ch;
	res = handle_operand(cur);

	return res;
}

static int store(char *cur)
{
	char *pos0, ch;
	int res;

	return_val_if_fail(cur_text_sect != NULL, -1);

	pos0 = cur;
	skip_id(&cur);

	ch = *cur;
	*cur = 0;

	if (STREQ("store", pos0)) {
		buffer_put_u8(cur_text_sect, STORE);
	} else if (STREQ("store0", pos0)) {
		buffer_put_u8(cur_text_sect, STORE0);
	} else if (STREQ("astore", pos0)) {
		buffer_put_u8(cur_text_sect, ASTORE);
	} else if (STREQ("astore0", pos0)) {
		buffer_put_u8(cur_text_sect, ASTORE0);
	} else if (STREQ("strstore", pos0)) {
		buffer_put_u8(cur_text_sect, STRSTORE);
	} else if (STREQ("strstore0", pos0)) {
		buffer_put_u8(cur_text_sect, STRSTORE0);
	} else if (STREQ("astrstore", pos0)) {
		buffer_put_u8(cur_text_sect, ASTRSTORE);
	} else if (STREQ("astrstore0", pos0)) {
		buffer_put_u8(cur_text_sect, ASTRSTORE0);
	} else {
		err_msg("unknown store instruction `%s'", pos0);
		return -1;
	}

	*cur = ch;
	res = handle_operand(cur);

	return res;
}

static int refer(char *cur)
{
	char *pos0, ch;
	int res;

	return_val_if_fail(cur_text_sect != NULL, -1);

	pos0 = cur;
	skip_id(&cur);

	ch = *cur;
	*cur = 0;

	if (STREQ("refcopy", pos0)) {
		buffer_put_u8(cur_text_sect, REFCOPY);
	} else if (STREQ("refput", pos0)) {
		buffer_put_u8(cur_text_sect, REFPUT);
	} else if (STREQ("refload", pos0)) {
		buffer_put_u8(cur_text_sect, REFLOAD);
	} else if (STREQ("refstore", pos0)) {
		buffer_put_u8(cur_text_sect, REFSTORE);
	} else if (STREQ("refstore0", pos0)) {
		buffer_put_u8(cur_text_sect, REFSTORE0);
	} else if (STREQ("refstrload", pos0)) {
		buffer_put_u8(cur_text_sect, REFSTRLOAD);
	} else if (STREQ("refstrstore", pos0)) {
		buffer_put_u8(cur_text_sect, REFSTRSTORE);
	} else if (STREQ("refstrstore0", pos0)) {
		buffer_put_u8(cur_text_sect, REFSTRSTORE0);
	} else {
		err_msg("unknown store instruction `%s'", pos0);
		return -1;
	}

	*cur = ch;
	res = handle_operand(cur);

	return res;
}

static int ainit(char *cur)
{
	int res;

	return_val_if_fail(cur_text_sect != NULL, -1);

	buffer_put_u8(cur_text_sect, AINIT);

	/* Skip instruction. */
	skip_id(&cur);
	skip_blanks(&cur);

	res = handle_operand(cur);

	return res;
}

/* Backpatch jumps with correct addresses. */
static void backpatch_labels()
{
	unsigned i, j;
	int ln1, ln2, patch;

	for (i = 0; i < need_patch.used; i++) {
		ln1 = need_patch.lb[i].ln;
		for (j = 0; j < labels.used; j++) {
			ln2 = labels.lb[j].ln;
			if (ln1 != ln2)
				continue;

			patch = labels.lb[j].addr - need_patch.lb[i].addr;
			buffer_put_u32_be_at(text_be_sect, patch,
							need_patch.lb[i].addr);
		}
	}
}

/* Convert a native format symbol table to a be format symbol table. */
static void convert_symbol_table_to_be()
{
	struct sym_tb_entry *sym_table;
	unsigned int size, i;

	sym_table = (struct sym_tb_entry *)buffer_ptr(symbol_sect);
	size = buffer_size(symbol_sect) / sizeof(struct sym_tb_entry);

	for (i = 0; i < size; i++) {
		struct sym_tb_entry *s = &sym_table[i];
		obj_put_sym_tb_entry_be(symbol_be_sect, s);
	}
}

/* Helpers to fill out headers. */
static void fill_obj_header(struct obj_header *h, unsigned int shnum, unsigned int shoff)
{
	strcpy((char *)h->o_magic, ".VM");
	h->o_type = FT_OBJ;
	h->o_maj_vers = MAJOR_VERS;
	h->o_min_vers = MINOR_VERS;
	h->o_shoff = shoff;
	h->o_shnum = shnum;
	h->o_shentsize = sizeof(struct section_header);
	h->o_entry = 0;
}

static void fill_text_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(string_sect, ".text");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = buffer_size(text_be_sect);
	sh->sh_type = SHT_TEXT;
}

static void fill_init_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(string_sect, ".init");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = buffer_size(init_be_sect);
	sh->sh_type = SHT_INIT;
}

static void fill_data_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(string_sect, ".data");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = data_section_size;
	sh->sh_type = SHT_DATA;
}

static void fill_ro_sect_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(string_sect, ".ro");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = buffer_size(ro_sect);
	sh->sh_type = SHT_RO;
}

static void fill_sym_tb_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(string_sect, ".symbol");
	sh->sh_entsize = sizeof(struct sym_tb_entry);
	sh->sh_off = off;
	sh->sh_size = buffer_size(symbol_be_sect);
	sh->sh_type = SHT_SYMTAB;
}

static void fill_rel_tb_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(string_sect, ".reloc");
	sh->sh_entsize = sizeof(struct rel_tb_entry);
	sh->sh_off = off;
	sh->sh_size = buffer_size(reloc_be_sect);
	sh->sh_type = SHT_RELTAB;
}

static void fill_string_sect_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(string_sect, ".string");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = buffer_size(string_sect);
	sh->sh_type = SHT_STRTAB;
}

/* Create a new file or if it already exists -- truncate it. */
static int open_file(const char *file_name)
{
	int fd;

	fd = open(file_name, O_CREAT|O_TRUNC|O_RDWR, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd < 0)
		xerror(1, "can't open file %s", file_name);

	return fd;
}

/* Gather all sections and headers in the final buffer and write it to the file
 * descriptor. */
static void write_file(const char *file_name)
{
	struct buffer *obj_file;
	struct obj_header obj_h;
	struct section_header text_sh, init_sh, data_sh, ro_sh;
	struct section_header sym_tb_sh, reloc_tb_sh, string_sh;
	unsigned int offs;
	int fd;

	obj_file = buffer_new();

	offs = sizeof(struct obj_header);
	fill_text_section_header(&text_sh, offs);

	offs += buffer_size(text_be_sect);
	fill_init_section_header(&init_sh, offs);

	offs += buffer_size(init_be_sect);
	fill_data_section_header(&data_sh, offs);
	fill_ro_sect_header(&ro_sh, offs);

	offs += buffer_size(ro_sect);
	fill_sym_tb_section_header(&sym_tb_sh, offs);

	offs += buffer_size(symbol_be_sect);
	fill_rel_tb_section_header(&reloc_tb_sh, offs);

	offs += buffer_size(reloc_be_sect);
	fill_string_sect_header(&string_sh, offs);
	offs += buffer_size(string_sect);

	fill_obj_header(&obj_h, 7, offs);
	obj_put_obj_header_be(obj_file, &obj_h);

	buffer_copy(obj_file, text_be_sect, buffer_size(text_be_sect));
	buffer_copy(obj_file, init_be_sect, buffer_size(init_be_sect));
	buffer_copy(obj_file, ro_sect, buffer_size(ro_sect));
	buffer_copy(obj_file, symbol_be_sect, buffer_size(symbol_be_sect));
	buffer_copy(obj_file, reloc_be_sect, buffer_size(reloc_be_sect));
	buffer_copy(obj_file, string_sect, buffer_size(string_sect));

	obj_put_section_header_be(obj_file, &text_sh);
	obj_put_section_header_be(obj_file, &init_sh);
	obj_put_section_header_be(obj_file, &data_sh);
	obj_put_section_header_be(obj_file, &ro_sh);
	obj_put_section_header_be(obj_file, &sym_tb_sh);
	obj_put_section_header_be(obj_file, &reloc_tb_sh);
	obj_put_section_header_be(obj_file, &string_sh);

	fd = open_file(file_name);

	buffer_write_fd(fd, obj_file);
	buffer_free(obj_file);
}

static void print_usage()
{
	printf("Usage: %s [options] file\n"
	" -h		print this help\n"
	" -o <file>	place the output into <file>\n", __progname);
}

/* Helpers to allocate and free resources. */
static void create_buffers()
{
	text_be_sect = buffer_new();
	init_be_sect = buffer_new();
	ro_sect = buffer_new();
	string_sect = buffer_new();
	reloc_be_sect = buffer_new();
	symbol_be_sect = buffer_new();
	symbol_sect = buffer_new();
}

static void release_buffers()
{
	buffer_free(text_be_sect);
	buffer_free(init_be_sect);
	buffer_free(ro_sect);
	buffer_free(string_sect);
	buffer_free(reloc_be_sect);
	buffer_free(symbol_be_sect);
	buffer_free(symbol_sect);
}

int main(int argc, char **argv)
{
	FILE *fp;
	char buf[1024];
	char *line;
	char *out_name = "file.vo";
	int opt, res;

	if (argc < 2) {
		print_usage();
		exit(1);
	}

	while ((opt = getopt(argc, argv, "ho:")) != -1) {
		switch (opt) {
		case 'h':
			print_usage();
			exit(0);
			break;
		case 'o':
			out_name = optarg;
			break;
		default:
			print_usage();
			exit(1);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if ((argc) != 1)
		err_msg("wait for one input file");

	file_name = argv[0];

	fp = fopen(file_name, "r");
	if (fp == NULL)
		xerror(1, "can't open file %s", file_name);

	create_buffers();
	/* Main cycle, read a file line by line. */
	while ((fgets(buf, sizeof(buf), fp)) != NULL) {
		size_t n;
		char *pos0, ch;
		int i;

		n = strlen(buf);

		if (buf[n-1] == '\n')
			buf[n-1] = 0;
		else
			continue;

		++line_number;

		line = buf;
		skip_blanks(&line);

		if (*line == 0)
			continue;

		if (*line == '#')
			continue;

		pos0 = line;

		if ((*pos0 == '.') || isalpha(*pos0)) {
			do {
				++pos0;
			} while (isalpha(*pos0));
		}
		/* Declaration has `:' at the end. */
		if (*pos0 == ':')
			++pos0;

		ch = *pos0;
		*pos0 = 0;
		/* Find a handler for this instruction. */
		for (i = 0; instr_handlers[i].name != NULL; i++) {
			if (STREQ(instr_handlers[i].name, line)) {
				*pos0 = ch;
				res = instr_handlers[i].handler(line);
				if (res < 0)
					goto err;
				break;
			}
		}
		/* It is not an instruction, it can be `name:' only. */
		if (instr_handlers[i].name == NULL) {
			if (!isalpha(*line)) {
				err_msg("unknown assembler expression %s", line);
				goto err;
			}
			*pos0 = ch;
			/* Call the last handler. */
			res = instr_handlers[i].handler(line);
			if (res < 0)
				goto err;
		}
	}
	/* Time to backpatch labels with correct addresses. */
	backpatch_labels();
	convert_symbol_table_to_be();

	write_file(out_name);

	release_buffers();
	fclose(fp);

	return 0;

err:
	release_buffers();
	fclose(fp);

	return 1;
}
